
/***************************************************************************
 *
 * ListBinaryHeap.java
 *
 * The ListBinaryHeap.java class implements a binary heap according to
 * the element class's implementation of the Comparable and Comparator interfaces
 *
 *****************************************************************************/

import java.util.*;


public class ListBinaryHeap<T extends Comparable<? super T>> implements BinaryHeap<T>
{
    private ArrayList<T> array; 		  // The heap ArrayList
    private Comparator<T> c = null;
    // Notice: we use a dummy node, currentSize = array.size()-1
    
    
    private boolean lessThan(int i, int j, Comparator<T> comparator){
  	if (comparator == null) return array.get(i).compareTo(array.get(j)) < 0;
  	else return comparator.compare(array.get(i), array.get(j)) < 0;
    }
    
    // initializer
    {  
        array = new ArrayList<T>();
        array.add(null);  // dummy element	
    }
    
    /**
     * Construct the binary heap.
     */
    public ListBinaryHeap( ) { }

    
    /**
     *  Constructs the binary heap given a comparator
     */
    public ListBinaryHeap(Comparator<T> comparator)
    {
    	c = comparator;
    }


	
    /**
      * Construct the binary heap given an array of items.
      */
     public ListBinaryHeap( T [ ] items)
     {
             for( T item : items ) array.add(item); 
             heapify( );
     }
     

     /**
      * Construct the binary heap given an array of items and a comparator.
      */
     public ListBinaryHeap( T [ ] items, Comparator<T> comparator)
     {
             c = comparator;
             for( T item : items ) array.add(item); 
             heapify( );
     }
     
  
     /**
      * @return the heap size 
      */
     public int size() {return array.size()-1;}

     /**
      * Test if the heap is logically empty.
      * @return true if empty, false otherwise.
      */
     public boolean isEmpty( )
     {
         return array.size() == 1; // it contains only the dummy element
     }
	
	/**
     * Insert into the ArrayList, maintaining heap order.
     * Duplicates are allowed.
     * @param x the item to insert.
     */
    public void insert( T x )
    {
        // Fix up
    	array.set(0,x);
    	array.add(x);
        int hole = array.size()-1;  //position of x
        while (lessThan(0, hole/2, c) ) {	
        	array.set(hole, array.get( hole/2 ));
        	hole /= 2; 
        }
        array.set(hole, x);    	
     }


	/**
     * Internal method to percolate down in the heap.
     * @param hole the index at which the percolate begins.
     */
    private void fixHeap( int hole )  // Fix down
    {
        int child, currentsize = array.size()-1;
        T tmp = array.get( hole );
      
        while (hole * 2 <= currentsize)
        {
            child = hole * 2;
            if( child != currentsize && lessThan(child + 1, child, c ) )
                child++;
            if( lessThan(child, hole, c) ) {
                array.set( hole, array.get(child) );
                hole = child;
                array.set(hole, tmp);
            }
            else break;
        }
        
    }

    
    /**
     * Establish heap order property from an arbitrary
     * arrangement of items. Runs in linear time.
     */
    public void heapify( )
    {   int currentsize = array.size()-1;
        for( int i = currentsize / 2; i > 0; i-- )
            fixHeap( i );  // Fix down
    }

  
    
    
    /**
     * Find the smallest item in the heap.
     * @return the smallest item, or throw an exception if empty.
     */
    public T findMin( )
    {
        if( isEmpty( ) )
		throw new NoSuchElementException("The binary heap is empty");
        return array.get(1);
    }

	
    /**
     * Remove the smallest item from the heap.
     * @return the smallest item, or throw an exception if empty.
     */
    public T deleteMin( )
    {
        if( isEmpty( ) )
        	throw new NoSuchElementException("The binary heap is empty");
        T minItem = array.get(1);  //findMin( );
        T lastItem = array.get(array.size()-1 ); 
        array.set( 1, lastItem );
        fixHeap( 1 );  // Fix down	
        lastItem = array.remove(array.size()-1 ); 
        return minItem;
    }



    /**
     * Make the heap logically empty.
     */
    public void clear( )
    {
        array = new ArrayList<T>();
        array.add(null);
    }
    



}
